#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Simple async crawler/callback queue based on gevent."""

import traceback
import logging
import httplib2

import gevent
from gevent import monkey, queue, event, pool
monkey.patch_all()

logger = logging.getLogger(__name__)

from aranha.job import Job

class Crawler(object):
    """Simple crawler based on gcrawler, but with a rolling inq that can be
    added to.  The crawler is done when the workers have no more jobs and
    there are no more urls in the queue."""
    seen_jobs = set()

    def __init__(self, handler, timeout=2, concurrency=4, pipeline_size=100):
        self.handler = handler
        self.handler.crawler = self
        self.timeout = timeout
        self.count = concurrency
        self.inq = queue.Queue(0)
        self.outq = queue.Queue(pipeline_size)
        self.jobq = queue.Queue()
        self.pool = pool.Pool(concurrency)
        self.worker_finished = event.Event()

        for job in getattr(self.handler, 'jobs', []):
            self.add_job(job)

        self.cache = '.cache'

    def start(self):
        """Start the crawler.  Starts the scheduler and pipeline first, then
        adds jobs to the pool and waits for the scheduler and pipeline to
        finish.  The scheduler itself shuts down the pool and waits on that,
        so it's not required to watch the pool."""
        # start the scheduler & the pipeline
        self.scheduler_greenlet = gevent.spawn(self.scheduler)
        self.pipeline_greenlet = gevent.spawn(self.pipeline)
        self.scheduler_greenlet.join()

    def add_job(self, job, **kwargs):
        """Add a job to the queue.  The job argument can either be a Job object
        or a url with keyword arguments to be passed to the Job constructor."""
        if isinstance(job, basestring):
            job = Job(job, **kwargs)
        # do not visit previously viewed urls
        if job in self.seen_jobs:
            return
        self.jobq.put(job)

    def scheduler(self):
        """A simple job scheduler.  Presuming it's started after there is at
        least one job, it feeds jobs to the job queue to a synchronous worker
        job queue one at a time.  When the worker queue fills up, the
        scheduler will block on the put() operation.  When job queue is empty,
        the scheduler will wait for the workers to finish.  If the job queue
        is empty and no workers are active, the pool's stopped."""
        logger = logging.getLogger(__name__ + '.scheduler')
        while True:
            # join dead greenlets
            for greenlet in list(self.pool):
                if greenlet.dead:
                    self.pool.discard(greenlet)
            try:
                logger.debug("Fetching job from job queue.")
                job = self.jobq.get_nowait()
            except queue.Empty:
                logger.debug("No jobs remaining.")
                if self.pool.free_count() != self.pool.size:
                    logger.debug("%d workers remaining, waiting..." % (self.pool.size - self.pool.free_count()))
                    self.worker_finished.wait()
                    self.worker_finished.clear()
                    continue
                else:
                    logger.debug("No workers left, shutting down.")
                    return self.shutdown()
            self.pool.spawn(self.worker, job)

    def shutdown(self):
        """Shutdown the crawler after the pool has finished."""
        self.pool.join()
        self.outq.put(StopIteration)
        self.pipeline_greenlet.join()
        return True

    def worker(self, job, logger=logging.getLogger(__name__ + '.worker')):
        """A simple worker that fetches urls based on jobs and puts the
        ammended jobs on the outq for processing in the pipeline thread.
        After each job is fetched, but before the worker sets the finished
        event, the handler's preprocess method will be called on the job. This
        is its opportunity to add urls to the job queue.  Heavy processing
        should be done via the pipeline in postprocess."""
        logger.debug("starting: %r" % job)
        # you need to create a new Http instance per greenlet, see ticket:
        #   http://code.google.com/p/httplib2/issues/detail?id=5
        h = httplib2.Http(self.cache)
        try:
            if job.body or job.method in ('PUT', 'POST'):
                job.response, job.data = h.request(job.url, method=job.method,
                    headers=job.headers, body=job.body)
            else:
                job.response, job.data = h.request(job.url, method=job.method,
                    headers=job.headers)
            (job.handler or self.handler).preprocess(job)
        except Exception, e:
            logger.error("Preprocessing error:\n%s" % traceback.format_exc())
        self.outq.put(job)
        self.worker_finished.set()
        logger.debug("finished: %r" % job)
        raise gevent.GreenletExit('success')

    def pipeline(self):
        """A post-processing pipeline greenlet which keeps post-processing from
        interfering with network wait parallelization of the worker pool."""
        logger = logging.getLogger(__name__ + '.pipeline')
        for job in self.outq:
            try:
                (job.handler or self.handler).postprocess(job)
            except:
                logger.error("Pipeline error:\n%s" % traceback.format_exc())
        logger.debug("finished processing.")

def run(handler, **kwargs):
    Crawler(handler, **kwargs).start()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Assorted utilities."""

import random
from aranha import VERSION

useragent = 'Aranha %s' % ('.'.join(map(str,VERSION)))

uas = (
    # webkit
    'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_4; en-US) '
        'AppleWebKit/533.4 (KHTML, like Gecko) Chrome/5.0.375.127 Safari/533.4',
    'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; en-us) '
        'AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27',
    # ffox
    'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0.1) Gecko/20100101 Firefox/4.0.1',
    # ies
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)',
    'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.0; WOW64; SLCC1; .NET '
        'CLR 2.0.50727; .NET CLR 3.0.04506; .NET CLR 1.1.4322)',
)

def ua(random=False):
    if random:
        return random.choice(uas)
    return useragent

def queueitems(queue):
    l = []
    while not queue.empty():
        l.append(queue.get_nowait())
    return l




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Aranha job encapsulation."""

import utils

class Job(object):
    """Encapsulation of a job to put in the job queue.  Contains at least
    a URL, but can also have custom headers or entirely custom meta
    information.  The response object and data are tacked onto this object,
    which is then returned to the handler for processing."""
    default_headers = {
        'Accept': 'Accept:text/html,application/xhtml+xml,application/xml;q=0.9,*/*;',
        'Accept-Encoding': 'gzip,deflate',
        'Cache-Control': 'max-age=0',
    }
    def __init__(self, url, headers=None, meta=None, method='GET', handler=None,
            body=None, randomua=False):
        self.__url = url
        self.method = method
        self.headers = dict(self.default_headers)
        self.headers.update(headers or {})
        self.headers['User-Agent'] = utils.ua(randomua)
        self.body = body
        self.meta = meta
        self.handler = handler

    def __hash__(self):
        return hash(self.url)

    def __cmp__(self, other):
        return cmp(hash(self), hash(other))

    def __repr__(self):
        return '<Job: %s (%s)>' % (
            self.url,
            'done: %d' % len(self.data) if hasattr(self, 'data') else 'pending',
        )

    url = property(lambda self: self.__url)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A simple interface to the base crawler and handlers that allows for natural
script-style usage of the spider.  The goal here is to provide convenience
functions to scale down the spidering task to the definition of a few callbacks
and the call of a function."""

import urllib

from aranha.utils import queueitems

from aranha.job import Job
from aranha.handler import BaseHandler, SimpleHandler
from aranha.crawler import Crawler

def job_from_request(request):
    """Create a job from a urllib2.Request object."""
    url = request.get_full_url()
    headers = dict(request.headers)
    method = request.get_method()
    body = request.data
    if isinstance(body, dict):
        body = urllib.urlencode(body)
    return Job(url, headers=headers, method=method, body=body)

def starturls(urls, callback, concurrency=4, timeout=2):
    """Fetch a number of urls.  These may be string urls or jobs."""
    handler = SimpleHandler(preprocess=callback)
    jobs = [Job(url, handler=handler) for url in urls]
    fetchjobs(jobs, concurrency, timeout, handler=handler)

def startjobs(jobs, concurrency=4, timeout=2, handler=None):
    """Fetch a number of jobs.  These jobs should have handlers set, as the
    BaseHandler with noop callbacks will be used."""
    handler = handler or BaseHandler()
    handler.jobs = jobs
    crawler = Crawler(handler, concurrency=concurrency, timeout=timeout)
    crawler.start()

def geturls(urls, concurrency=4, timeout=5):
    """Get a list of urls.  This synchronous call fetches all urls
    asynchronously and returns the data within."""
    from gevent import queue
    q = queue.Queue()
    def callback(job, crawler):
        q.put({'url': job.url, 'document': job.data})
    handler = SimpleHandler(preprocess=callback)
    jobs = [Job(url, handler=handler) for url in urls]
    handler.jobs = jobs
    crawler = Crawler(handler, concurrency=concurrency, timeout=timeout)
    crawler.start()
    return queueitems(q)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Handler classes designed to work with the crawler."""

class BaseHandler(object):
    """Base handler object that does nothing."""

    def preprocess(self, job):
        pass

    def postprocess(self, job):
        pass

class SimpleHandler(BaseHandler):
    """A simple handler that takes pre and post processing methods as
    arguments.  The crawler will be passed to these callbacks as a crawler
    kwarg."""
    def __init__(self, preprocess=None, postprocess=None):
        self.pre = preprocess
        self.post = postprocess
        super(SimpleHandler, self).__init__()

    def preprocess(self, job):
        if self.pre:
            self.pre(job, crawler=self.crawler)

    def postprocess(self, job):
        if self.post:
            self.post(job, crawler=self.crawler)

